#! python
###########################################################################
# 
# File:            fms_rka.py
#
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Not allowed to be used for mass downloading
#                  of maps!
#
#
HELP_TEXT = """
# Module options:      --maastokartta - Use maastokartta as background map
#                                       (default)
#
#                      --ilmakuva     - Use aerial pictures as background map
#
#                      --rinnevarjostus - Use hillshade ("3D") pictures as background map
#
#                      --taustakartta - Use taustakartta material as background map
#
#                      --selkokartta  - Use selkokartta material as background map
#
"""
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         3.3b
#
# Date:            31.01.2022
#
# Functions:       -
#                   
# Description:     Map service class for Retkikartta.fi current material.
#
# Requirements:    Python interpreter 3.6 or newer (www.python.org)
#                  (tested with 3.10.1)
# 
#                  Python Imaging Library (tested with 1.1.5) 
#                  (www.pythonware.org/products/pil/)
#
# Version history: ** 01.10.2013 v3.2a (Olli Lammi) **
#                  First implementation for 2013 material.
#
#                  ** 05.12.2016 v3.2e (Olli Lammi) **
#                  Fixed non working Retkikartta.fi.
#
#                  ** 17.01.2033 v3.3a (Olli Lammi) **
#                  Python3 version.
#
#                  ** 31.01.2033 v3.3b (Olli Lammi) **
#                  Fixed service error handling. Added cookie support.
#
###########################################################################

# Imports

from doctest import register_optionflag
import sys, os
import math, re, os.path
import tempfile, shutil

import urllib.request
import urllib.parse
import http.cookiejar

from PIL import Image  # Python Imaging Library 

import coordinates  # KKJ to and from WGS84 coordinate transforms

import fms


###########################################################################

# Constants

MOD_VERSION = 'v3.3a'
SERVICE_CODE = 'rka'
CACHE_CODE = 'rka202201'

# constants for the picture and distance sizes and ratios
PIC_ORIG_SIZEX = 256
PIC_ORIG_SIZEY = 256
PIC_SIZEX = PIC_ORIG_SIZEX
PIC_SIZEY = PIC_ORIG_SIZEY
TILE_ORIGIN_X = -548576.0
TILE_ORIGIN_Y = 8388608.0

PIC_SCALES = { '1:1000': [15, 0.25, 'JHS180 level 15'], \
               '1:2000': [14, 0.5, 'JHS180 level 14'], \
               '1:4000': [13, 1.0, 'JHS180 level 13'], \
               '1:8000': [12, 2.0, 'JHS180 level 12'], \
               '1:16000': [11, 4.0, 'JHS180 level 11'], \
               '1:32000': [10, 8.0, 'JHS180 level 10'], \
               '1:64000': [9, 16.0, 'JHS180 level 9'], \
               '1:128000': [8, 32.0, 'JHS180 level 8'],\
               '1:256000': [7, 64.0, 'JHS180 level 7'],\
               '1:512000': [6, 128.0, 'JHS180 level 6'],\
               '1:1024000': [5, 256.0, 'JHS180 level 5'],\
               '1:2048000': [4, 512.0, 'JHS180 level 4'],\
               '1:4096000': [3, 1024.0, 'JHS180 level 3'] }

PIC_SCALE_DEFAULT = '1:16000'

PIC_SERVICES = { \
               'maastokartta': {'layer': 'maastokartta', 'ext': 'png'}, \
               'ilmakuva': {'layer': 'ortokuva', 'ext': 'jpeg'}, \
               'rinnevarjostus': {'layer': 'korkeusmalli_vinovalo', 'ext': 'png'}, \
               'taustakartta': {'layer': 'taustakartta', 'ext': 'png'}, \
               'selkokartta': {'layer': 'selkokartta', 'ext': 'png'} \
               }
PIC_SERVICE_DEFAULT = 'maastokartta'

LAYERS = { 'kesareitti': ['Kesäretkeilyreitit', 'summer-hiking-trail', 1.0], \
           'luontopolku': ['Luontopolku', 'nature-trail', 1.0], \
           'talvireitti': ['Talviretkeilyreitti', 'winter-excursion-route', 1.0], \
           'latupertsa': ['Latu, perinteinen hiihtotapa', 'ski-trail-traditional-skiing-style', 1.0], \
           'latuvapaa': ['Latu, vapaa hiihtotapa', 'ski-trail-free-skiing-style', 1.0], \
           'pyorareitti': ['Pyöräilyreitti', 'bicycle-route', 1.0], \
           'maastopyorareitti': ['Maastopyöräilyreitti', 'mountain-biking-route', 1.0], \
           'vesiretkeilyreitti': ['Vesiretkeilyreitti', 'canoe-route', 1.0], \
           'veneilyreitti': ['Veneilyreitti', 'boating-route', 1.0], \
           'ulkoilutie': ['Ulkoilutie', 'recreation-way', 1.0], \
           'kuntorata': ['Kuntorata', 'fitness-trail', 1.0], \
           'esteetonreitti': ['Esteetön reitti', 'accessible-route', 1.0], \
           'kelkkareitti': ['Moottorikelkkailureitti', 'snowmobile-route', 1.0], \
           'kelkkaura': ['Moottorikelkkaura', 'snowmobile-track', 1.0], \
           'kansallispuisto': ['Kansallispuistot', 'national-parks', 0.4], \
           'retkeilyalue': ['Retkeilyalueet', 'hiking-areas', 0.4], \
           'luonnonpuisto': ['Luonnonpuistot', 'strict-nature-reserves', 0.4], \
           'suojelualue': ['Muut suojelualueet', 'other-protected-areas', 0.4], \
           'lapineramaa': ['Lapin erämaa-alueet', 'laplands-wilderness-areas', 0.4], \
           'vapaalue': ['Vapa-alueet - kalastus', 'rod-fishing', 0.5], \
           'pyydysalue': ['Pyydysalueet - kalastus', 'trap-fishing', 0.5], \
           'pienriista': ['Pienriista-alueet - metsästys', 'small-game-areas', 0.5], \
           'hirvialue': ['Hirvialueet - metsästys', 'moose-hunting-areas', 0.5], \
           'karhualue': ['Karhunmetsästys poronhoitoalueella', 'bear-hunting-in-the-reindeer-herding-region', 0.5], \
           'monikayttometsa': ['Monikäyttömetsät', 'multiple-use-forests', 0.4], \
           'luontokohde': ['Luontokohteet', 'sites-of-high-natural-value', 0.4], \
           'tutkimusmetsa': ['Tutkimus- ja opetusmetsät', 'research-and-educational-forests', 0.4], \
           'muuerityis': ['Muut erityiskohteet', 'other-special-sites', 0.4], \
           'ylalapinluontaistalous': ['Ylä-Lapin luontaistalousalueet', 'natural-economy-areas-in-northern-lapland', 0.4], \
           'kuntaulkoilu': ['Kuntien ulkoilualueet', 'outdoor-areas-of-municipalities', 0.4], \
           'muualue': ['Muut alueet', 'other-areas', 0.4], \
           'yleinenvesialue': ['Yleiset vesialueet', 'public-waters', 0.4], \
           'syvyys': ['Syvyyskäyrät', 'depth-contours', 1.0] }

LAYERORDER = ['kesareitti', 'luontopolku', 'talvireitti', 'latupertsa', 'latuvapaa', 'pyorareitti', 'maastopyorareitti', \
              'vesiretkeilyreitti', 'veneilyreitti', 'ulkoilutie', 'kuntorata', 'esteetonreitti', 'kelkkareitti', 'kelkkaura', \
              'kansallispuisto', 'retkeilyalue', 'luonnonpuisto', 'suojelualue', 'lapineramaa', 'vapaalue', 'pyydysalue', \
              'pienriista', 'hirvialue', 'karhualue', 'monikayttometsa', 'luontokohde', 'tutkimusmetsa', 'muuerityis', \
              'ylalapinluontaistalous', 'kuntaulkoilu', 'muualue', 'yleinenvesialue', 'syvyys']


###########################################################################

# Module globals

COOKIES = http.cookiejar.CookieJar()

###########################################################################

# Classes

class FMSRetkikarttaFi(fms.FetchMapService):
    def __init__(self):
        self.SERVICE_CODE = SERVICE_CODE
        self.CACHE_CODE = CACHE_CODE

        self.PIC_SERVICE = PIC_SERVICE_DEFAULT

        self.PIC_SCALE_DEFAULT = PIC_SCALE_DEFAULT
        self.SetScale(self.PIC_SCALE_DEFAULT)
     
    
    ###########################################################################
    # Function:    GetVersion
    ###########################################################################
    # Description: Returns the module version string. 
    ###########################################################################    
    def GetVersion(self):
        return MOD_VERSION
    
    
    ###########################################################################
    # Function:    HandleParameter
    ###########################################################################
    # Description: Handles command line parameter. Returns 0, if the parameter
    #              was invalid. Returns 1 if the parameter was ok. 
    ###########################################################################  
    def HandleParameter(self, param):
        if len(param) > 2 and param[:2] == '--':
            temp = param[2:]
            if temp in PIC_SERVICES:
                self.PIC_SERVICE = temp
                return 1
            else:
                return 0
        return 0


    ###########################################################################
    # Function:    SetScale
    ###########################################################################
    # Description: Initializes the service and values to retrieve map parts
    #              in given scale. The parameter must be one of the values
    #              returned by GetScales.
    ###########################################################################
    def SetScale(self, scale):
        # determine picture scale and ratio
        try:
            self.PIC_SIZEX = PIC_SIZEX
            self.PIC_SIZEY = PIC_SIZEY
            self.SCALE = scale
            self.SCALEVAL = PIC_SCALES[scale][0]
            self.PIC_RATIO = PIC_SCALES[scale][1]
            self.MAP_SIZEX = self.PIC_SIZEX * self.PIC_RATIO
            self.MAP_SIZEY = self.PIC_SIZEY * self.PIC_RATIO
        except:
            return 0
        return 1


    ###########################################################################
    # Function:    GetScales
    ###########################################################################
    # Description: Returns list of scales that the service supports
    ###########################################################################    
    def GetScales(self):
        scales = []
        for sc in PIC_SCALES.keys():
            scales.append([PIC_SCALES[sc], sc])
        scales.sort()
        scales.reverse()
        scales2 = []
        for sc in scales:
            scales2.append([sc[1], sc[0][2], sc[0][1]])
        return scales2
    

    ###########################################################################
    # Function:    GetPicIndexes
    ###########################################################################
    # Description: Returns two (x,y) tuples with indexes for map image that has
    #              the given (mapx, mapy) point and the lower left xy
    #              coordinates of the indexed map picture.
    ###########################################################################
    def GetPicIndexes(self, mapx, mapy):
        tilex = int(math.floor((mapx + (0.25 * self.PIC_RATIO) - TILE_ORIGIN_X) / self.MAP_SIZEX))
        tiley = int(math.floor((TILE_ORIGIN_Y - (mapy + (0.25 * self.PIC_RATIO))) / self.MAP_SIZEY))

        (fetchx, fetchy) = self.GetTileLowerLeft(tilex, tiley)
                                    
        return ((tilex, tiley), (fetchx, fetchy))

    def GetTileLowerLeft(self, tilex, tiley):
        llx = TILE_ORIGIN_X + tilex * self.MAP_SIZEX
        lly = TILE_ORIGIN_Y - (tiley + 1) * self.MAP_SIZEY
        return (llx, lly)
    

    ###########################################################################
    # Method:      FetchPic
    ###########################################################################
    # Description: Fetches one map tile from Retkikartta.fi.
    #              Result file will saved in cachedir. Tempdir is used as temp directory  
    #              while downloading.
    ###########################################################################

    def FetchPic(self, fetchr, fetchc, cachedir, tempdir, layer):
        fname = cachedir + self.GetCacheName(fetchc, fetchr, layer)

        tempfname = None
        if len(layer) > 0:
            tempfname = self.FetchPicLayer(fetchr, fetchc, tempdir, layer)
        else:            
            waitsize_x = PIC_ORIG_SIZEX
            waitsize_y = PIC_ORIG_SIZEY

            service = PIC_SERVICES[self.PIC_SERVICE]
            ext = service['ext']
            zoomlevel = self.SCALEVAL
            tile_x = fetchc
            tile_y = fetchr
            url = "https://maps-mml.anderscloud.com/wmts/?ngsw-bypass=1&layer=%s&style=default&tilematrixset=ETRS-TM35FIN&Service=WMTS&Request=GetTile&Version=1.0.0&Format=image%%2F%s&TileMatrix=%s&TileCol=%d&TileRow=%d" % (service['layer'], service['ext'], zoomlevel, tile_x, tile_y)
            USERAGENT = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.99 Safari/537.36"

            res = ''
            try:
              opener = urllib.request.build_opener(urllib.request.HTTPCookieProcessor(COOKIES))
              req = urllib.request.Request(url)
              req.add_header("Referrer-Policy", 'strict-origin-when-cross-origin')
              req.add_header("accept", 'image/avif,image/webp,image/apng,image/svg+xml,image/*,*/*;q=0.8')
              req.add_header("accept-encoding", 'gzip, deflate, br')
              req.add_header("accept-language", 'en-US,en;q=0.9')
              req.add_header("referer", 'https://retkikartta.fi/')
              req.add_header('sec-ch-ua', '" Not;A Brand";v="99", "Google Chrome";v="97", "Chromium";v="97"')
              req.add_header("sec-ch-ua-mobile", '?0')
              req.add_header("sec-ch-ua-platform", "Linux")
              req.add_header("sec-fetch-dest", 'image')
              req.add_header("sec-fetch-mode", 'no-cors')
              req.add_header("sec-fetch-site", 'cross-site')
              req.add_header("user-agent", USERAGENT)

              f = opener.open(req)
              res = f.read()
              contenttype = f.getheader('Content-Type')
              if contenttype and contenttype[:5].lower() != 'image':
                  print("ERROR: Invalid map image content type: %s" % contenttype)
                  f.close()
                  return 0
              f.close()
            except:
              print(sys.exc_info())
              return 0

            tempfname = ''
            # create a temporary file name for the map
            try:
                (tempfd, tempfname) = tempfile.mkstemp('.' + ext, 'tmpmap_', tempdir)
                os.close(tempfd)
            except:
                print('ERROR: Cannot create temporary file for map image file')
                return 0
    
            try:
                f = open(tempfname,'wb')
                f.write(res)
                f.close()
            except:
                print("ERROR: Cannot write temporary map image file: " + tempfname)
                return 0
        
            try:
                im = Image.open(tempfname)
                im.load()
        
                if im.size[0] != waitsize_x or im.size[1] != waitsize_y:
                    print("ERROR: Retrieved image is of invalid size (received: %dx%d, expected: %dx%d)" % (im.size[0], im.size[1], waitsize_x, waitsize_y))
                    try:
                        os.unlink(tempfname)
                    except:
                        pass
                    return 0 
                        
                im.save(tempfname)
            except:
                print("ERROR: Cannot convert map image or rewrite temporary map image file: " + tempfname)
                print(sys.exc_info())
                try:
                    os.unlink(tempfname)
                except:
                    pass
                return 0

        if tempfname == None:
            return 0
 
        # check if the map cache hash directory is created
        dname = os.path.dirname(fname)
        if not os.path.isdir(dname):
            try:
                os.makedirs(dname)
            except:
                print('ERROR: Cannot create map image directory: ' + dname)
                try:
                    os.unlink(tempfname)
                except:
                    pass
                return 0
    
        # move the temporary map file to the cache
        try:
            shutil.move(tempfname, fname)  
        except:
            print('ERROR: Cannot move loaded temporary map image to map cache.')
            try:
                os.unlink(tempfname)
            except:
                pass
            return 0
  
        return 1


    def FetchPicLayer(self, fetchr, fetchc, tempdir, layer):
        LAYER_PIC_SIZES = [ (PIC_ORIG_SIZEX, PIC_ORIG_SIZEY), \
                            (3 * PIC_ORIG_SIZEX, 3 * PIC_ORIG_SIZEY), \
                            (5 * PIC_ORIG_SIZEX, 5 * PIC_ORIG_SIZEY), \
                            (7 * PIC_ORIG_SIZEX, 7 * PIC_ORIG_SIZEY) ]

        tempfname = None
        for lps in LAYER_PIC_SIZES:                    
            waitsize_x = lps[0]
            waitsize_y = lps[1]

            margin_x = (waitsize_x - PIC_ORIG_SIZEX) / 2.0
            margin_y = (waitsize_y - PIC_ORIG_SIZEY) / 2.0
            cropbox = (int(margin_x), int(margin_y), int(margin_x) + PIC_ORIG_SIZEX, int(margin_y) + PIC_ORIG_SIZEY)

            (fetchx_ll, fetchy_ll) = self.GetTileLowerLeft(fetchc, fetchr)
            fetchx_ll = fetchx_ll - self.PIC_RATIO * margin_x
            fetchy_ll = fetchy_ll - self.PIC_RATIO * margin_y
            fetchx_ur = fetchx_ll + self.PIC_RATIO * waitsize_x
            fetchy_ur = fetchy_ll + self.PIC_RATIO * waitsize_y

            url = 'https://retkikartta-map.anderscloud.com/wms/'
            url = url + '?SERVICE=WMS&VERSION=1.3.0&REQUEST=GetMap&TRANSPARENT=true&ngsw-bypass=1'
            url = url + '&FORMAT=image%2Fpng'
            url = url + "&LAYERS=" + urllib.parse.quote(LAYERS[layer][1])
            url = url + '&CRS=EPSG%3A3067&STYLES='

            url = url + '&WIDTH=%d&HEIGHT=%d' % (waitsize_x, waitsize_y)

            tempval = "%f,%f,%f,%f" % (fetchx_ll, fetchy_ll, fetchx_ur, fetchy_ur)
            url = url + '&BBOX=' + urllib.parse.quote(tempval)
    
            USERAGENT = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.99 Safari/537.36"

            res = ''
            try:
                opener = urllib.request.build_opener(urllib.request.HTTPCookieProcessor(COOKIES))
                req = urllib.request.Request(url)
                req.add_header("Referrer-Policy", 'strict-origin-when-cross-origin')
                req.add_header("accept", 'image/avif,image/webp,image/apng,image/svg+xml,image/*,*/*;q=0.8')
                req.add_header("accept-encoding", 'gzip, deflate, br')
                req.add_header("accept-language", 'en-US,en;q=0.9')
                req.add_header("referer", 'https://retkikartta.fi/')
                req.add_header('sec-ch-ua', '" Not;A Brand";v="99", "Google Chrome";v="97", "Chromium";v="97"')
                req.add_header("sec-ch-ua-mobile", '?0')
                req.add_header("sec-ch-ua-platform", "Linux")
                req.add_header("sec-fetch-dest", 'image')
                req.add_header("sec-fetch-mode", 'no-cors')
                req.add_header("sec-fetch-site", 'cross-site')
                req.add_header("user-agent", USERAGENT)
      
                f = opener.open(req)
                res = f.read()
                contenttype = f.getheader('Content-Type')
                if contenttype and contenttype[:5].lower() != 'image':
                    print("ERROR: Invalid map image content type: %s" % contenttype)
                    f.close()
                    continue
                f.close()
            except:
                continue
       
            tempfname = ''
            # create a temporary file name for the map
            try:
                (tempfd, tempfname) = tempfile.mkstemp('.png', 'tmpmap_', tempdir)
                os.close(tempfd)
            except:
                print('ERROR: Cannot create temporary file for map image file')
                return None
    
            try:
                f = open(tempfname,'wb')
                f.write(res)
                f.close()
            except:
                print("ERROR: Cannot write temporary map image file: " + tempfname)
                return None
        
            try:
                im = Image.open(tempfname)
                im.load()
        
                if im.size[0] != waitsize_x or im.size[1] != waitsize_y:
                    print("ERROR: Retrieved image is of invalid size (received: %dx%d, expected: %dx%d)" % (im.size[0], im.size[1], waitsize_x, waitsize_y))
                    try:
                        os.unlink(tempfname)
                    except:
                        pass
                    tempfname = None
                    continue
    
                im = im.crop(cropbox)    
                im.save(tempfname)
            except Exception as e:
                try:
                    os.unlink(tempfname)
                except:
                    pass
                tempfname = None
                continue
            
            break
            
        if tempfname == None:
            print("ERROR: Cannot fetch map layer image.")
        return tempfname
 
         
         
    ###########################################################################
    # Method:    GetCacheName
    ###########################################################################
    # Description: Generates a name for the map image in cache according to
    #              location and other parameters.
    ###########################################################################
    def GetCacheName(self, x, y, layer):
        fname = "%s_%d_%d" % (self.SCALEVAL, int(self.PIC_SIZEX), int(self.PIC_SIZEY))
  
        ext = 'png'
        if len(layer) > 0:
            fname = fname + '_' + layer
            ext = 'png'
        elif self.PIC_SERVICE != PIC_SERVICE_DEFAULT:
            fname = fname + '_' + self.PIC_SERVICE
            ext = PIC_SERVICES[self.PIC_SERVICE]['ext']

        (hashx, hashy) = self.GetTileLowerLeft(x, y)
        hash = "%d_%d" % (HashRound(hashy), HashRound(hashx))
        fname = fname + '/' + hash
  
        fname = fname + "/map_%s_%s" % (self.CACHE_CODE, self.SCALEVAL)  
        
        fname = fname + '_'+ repr(int(round(y))) + '_' + repr(int(round(x))) + '.' + ext

        return fname


    ###########################################################################
    # Method:      PostProcess
    ###########################################################################
    # Description: Post processes the entire constructed map image according
    #              to service requirements.
    ###########################################################################    
    def PostProcess(self, im):
        pass


    ###########################################################################
    # Method:      TranslateToServiceCoordinatesXY
    ###########################################################################
    # Description: Does service require coordinates to be transferred to YKJ?
    #              Returns: tuple (type, x, y) where
    #              type - type string of the coordinate system
    #              x - easting or lon coordinate 
    #              y - northing or lat coordinate
    ###########################################################################       
    def TranslateToServiceCoordinatesXY(self, type, x, y):
        if (type != coordinates.COORD_TYPE_ETRSTM35FIN):
            temp = coordinates.Translate({'type': type, 'N': y, 'E': x}, \
                                         coordinates.COORD_TYPE_ETRSTM35FIN)
            return (temp['type'], temp['E'], temp['N'])
        return (type, x, y)

    
    ###########################################################################
    # Method:      GetHelp
    ###########################################################################
    # Description: Get service module specific help string.
    ###########################################################################    
    def GetHelp(self):
        return HELP_TEXT


    ###########################################################################
    # Method:      GetServiceDelayLimits
    ###########################################################################
    # Description: Get service specific delay times as tuple (min, max) in sec.
    ###########################################################################               
    def GetServiceDelayLimits(self):
        return (2, 6)
        
        
    ###########################################################################
    # Method:      GetAvailableLayers
    ###########################################################################
    # Description: Get list of available layer names. 
    ###########################################################################               
    def GetAvailableLayers(self):
        return LAYERORDER


    ###########################################################################
    # Method:      GetLayerInfo
    ###########################################################################
    # Description: Get info for given layer (list of layer name and opacity).  
    ###########################################################################               
    def GetLayerInfo(self, layer):
        return [LAYERS[layer][0], LAYERS[layer][2]]

    ###########################################################################
    # Method:      IsSelfCaching
    ###########################################################################
    # Description: Should return true, if module is self caching and
    #              GetCachePic will provide the image from the cache. 
    ###########################################################################   
    def IsSelfCaching(self):
        return 0
        
        
###########################################################################

# Functions

def HashRound(x):
    return ((x / 10000) / 5) * 5 

